/**
 * The engine loads, initializes, and executes the game.  Note that at the very very bottom of this file the "e" variable is instanced
 * and placed in the global namespace!  This will overwrite any other "e" variables we may have incidentally created.
 * @author Torgie
 */

/**
 * The engine has the following dependencies
 */
document.write('<script type="text/javascript" src="engine/gameState.js"></script>');
document.write('<script type="text/javascript" src="engine/scene.js"></script>');
document.write('<script type="text/javascript" src="engine/sprite.js"></script>');

(function(window) {

function engine() {

	// Create a reference to myself in the "engine" namespace for our setInterval calls.
	// Also used during mapKeyEvent calls which need the correct scope.
	// See brothercake's reply here: http://www.codingforums.com/archive/index.php/t-49047.html
	var self = this;

	// Store exactly when the engine was started
	this.initTimestamp = new Date();

	/**
	 * Starts the engine. Is responsible for initial setup and creation of game assets.
	 * This is usually the first function call on the engine.
	 * Called by: manually
	 */
	this.go = function(p) {
		// Debugging level: false = off, true = on
		this.debugging = (typeof p.debug == "boolean") ? p.debug : false;
		this.debug("Engine started!");

		// Create and store the engine gameState
		this.gs = new gameState(p.canvas);

		// Get our graphical canvas ready
		this.canvas = $(p.canvas)[0];
		this.context = this.canvas.getContext("2d");
		this.context.fillStyle = "#000";

		// Declare and execute the scene files
		this.declareScene(p.scene);
		// Calculate the total number of resources we will need to load
		for(var i in this.resources)
			this.totalResources++;

		this.debug("Scene loaded: " + this.totalSprites + " sprite" + (this.totalSprites != 1 ? 's' : '') + " and " + this.totalResources + " resource" + (this.totalResources != 1 ? 's' : ''));

		// Run the initLoop and load up the resources
		this.initLoop = setInterval(function(){
			self.init();
		}, 100);
	};//end function go

	this.sceneSettings = {};

	/**
	 * Takes in a scene file (like scenes/world1.js) and batch processes the sprites therein.
	 * Called by: go
	 */
	this.declareScene = function(scene) {
		$(this.canvas).css({
			background: "#000"
		});

		this.sceneSettings = scene.settings;

		this.gs.cameraMinX = this.sceneSettings.cameraMinX;
		this.gs.cameraMaxX = this.sceneSettings.cameraMaxX;
		this.gs.cameraMinY = this.sceneSettings.cameraMinY;
		this.gs.cameraMaxY = this.sceneSettings.cameraMaxY;

		// Place every sprite into the scene.
		for(var i in scene.sprites)
			for(var j in scene.sprites[i])
				this.declareSprite(i, scene.sprites[i][j]);

		// Go through every sprite that's remotely involved by the scene. This will skip entries already caught by declareSprite above
		for(var i in window)
			if(i.match(/sprite_/) != null)
				this.initResource(new window[i]({}));

		this.totalSprites = this.countSprites();
		this.totalAudio = this.countAudio();
	};//end function executeScene


	// Holds every sprite in the scene, by layer
	this.sprites = {
		background5: [],
		background4: [],
		background3: [],
		background2: [],
		background1: [],
		game: [],
		foreground: [],
		hud: []
	};
	this.nextSpriteID = 1; // Each sprite is given a unique ID, starting at 1

	// Holds every resource in the scene by name. Many sprites may reference one resource, and many resources may be used by one sprite. 
	this.resources = {};
	this.totalResources = 0;

	// Holds every audio resource in the scene by name. Work very much like image resources. (see above)
	this.audio = {};
	this.totalAudio = 0;
	this.audioReady = 0;

	/**
	 * This is the workhorse function for pushing a sprite into the game. ALL sprites must use this function to be placed.
	 * Called by: declareScene, manually
	 */
	this.declareSprite = function(layer, sprite) {
		// Increment the ID and throw it on the appropriate stack
		sprite.id = (typeof sprite.uniqueID != "undefined") ? sprite.uniqueID : this.nextSpriteID++;

		if (layer == "background5")
			sprite.speedChop = (0);
		else if (layer == "background4")
			sprite.speedChop = (1/8);
		else if (layer == "background3")
			sprite.speedChop = (1/4);
		else if (layer == "background2")
			sprite.speedChop = (1/2);

		this.sprites[layer].push(sprite);
		this.initResource(sprite);

		return sprite.id;
	};//end function declareSprite

	// Stores all of the intervals created by setInterval inside the engine
	this.intervals = [];
	this.declareInterval = function(callback, timeout) {
		this.intervals.push({
			callback: callback,
			timeout: timeout,
			interval: setInterval(callback, timeout)
		});
	};//end function declareInterval

	this.initResource = function(sprite) {
		/**
		 * Deal with the graphic resource(s) for this sprite. If it's a list of resources, add them all as <name>_<srcLabel>
		 */
		if (typeof sprite.src == "object") {
			for (var i in sprite.src) {
				if (typeof this.resources[sprite.name + "_" + i] == "undefined") {
					this.resources[sprite.name + "_" + i] = new Image();
					this.resources[sprite.name + "_" + i].src = sprite.src[i];
				}
			}
		}
		// Otherwise add it as just <name>
		else {
			if (typeof this.resources[sprite.name] == "undefined") {
				this.resources[sprite.name] = new Image();
				this.resources[sprite.name].src = sprite.src;
			}
		}

		/**
		 * Deal with the audio resource(s) for this sprite
		 */
		if (typeof sprite.audio == "object") {
			for (var i in sprite.audio) {
				if (typeof this.audio[sprite.name + "_" + i] == "undefined") {

					// Create out audio stream
					this.audio[sprite.name + "_" + i] = new Audio();

					// Make a local reference to the audio object and set the listener for the canplaythrough event handler 
					var a = this.audio[sprite.name + "_" + i];
					$(a).bind("canplaythrough", function() {
						self.audioReady++;
					});

					// Set the volume if a default is given
					if(typeof sprite.audio[i].volume != "undefined")
						a.volume = sprite.audio[i].volume;

					// What kind of audio can we play?
					// Currently canPlayType(type) returns: "", "maybe" or "probably" 
					if (a.canPlayType) {
						var canPlayMp3 = !!a.canPlayType && "" != a.canPlayType("audio/mpeg");
						var canPlayWav = !!a.canPlayType && "" != a.canPlayType("audio/x-wav");
						var canPlayOgg = !!a.canPlayType && "" != a.canPlayType('audio/ogg; codecs="vorbis"');
					}

					// Audio settings go here
					a.loop = (typeof sprite.audio[i].loop == "boolean") ? sprite.audio[i].loop : false;

					// Loop through each potential audi source
					for (var j in sprite.audio[i].src) {
						// If we've found a source, bail on this loop
						if(a.src != "")
							continue;

						// Figure out what extension the audio file is in
						var extension = sprite.audio[i].src[j].match(/.([a-z0-9]+)$/);

						if (extension == null) {
							console.debug(extension);
							throw "Could not determine extension type on the file: " + sprite.audio[i].src[j];
						}
						else 
							extension = extension[1];

						// In order of preference, assign the audio source type
						if(extension == "ogg" && canPlayOgg)
							a.src = sprite.audio[i].src[j];
						else if (extension == "mp3" && canPlayMp3)
							a.src = sprite.audio[i].src[j];
						else if (extension == "wav" && canPlayWav)
							a.src = sprite.audio[i].src[j];
						else
							continue;
					}

					// If we still don't have a source, just die, man.. it's over.
					if(a.src == "")
						throw "None of the audio objects supplied in " + sprite.name + " were valid";

					// Firefox absolutely, really, truly wants this to be here. Chrome will obey the preload defaults
					a.load();
				}
			}
		}
	};

	this.getSprite = function(spriteID) {
		for (var i in this.sprites) {
			for (var j in this.sprites[i]) {
				if (this.sprites[i][j].id == spriteID)
					return this.sprites[i][j];
			}
		}
	};//end function getSprite
	
	this.switchSpriteLayer = function(spriteId, fromLayer, toLayer) {
	    for (var i in this.layers.fromLayer) {
	        var spriteId = this.layers.fromLayer[i].id;
	        if (sprite.id == spriteId) {
//	        	this.layers.fromLayer[i] = void(0);
			this.fromLayer.splice(i, 1);
			this.layers.toLayer.push(sprite);
	        break;
	        }
	    }
	};

	this.destroySprite = function(spriteID, instant) {
		for (var i in this.sprites) {
			for (var j in this.sprites[i]) {
				// If we're trying to instantly nuke a sprite, do so now
				if (instant && this.sprites[i][j].id == spriteID) {
					this.sprites[i].splice(j, 1);
					return;
				}

				// Otherwise, run the destroy function on the sprite for fun effects
				if (this.sprites[i][j] != null && this.sprites[i][j].id == spriteID)
					this.sprites[i][j].destroy(this, i, j);

				// If the sprite is done dying (null, or instantly killed), remove it from the scene graph
				if(this.sprites[i][j] == null)
					this.sprites[i].splice(j, 1);
			}
		}
	};//end function destroySpite

	this.countSprites = function() {
		var tally = 0;
		for(var i in this.sprites)
			tally += this.sprites[i].length;
		return tally;
	};//end function countSprites

	this.countAudio = function() {
		var tally = 0;
		for(var i in this.audio)
			tally++;
		return tally;
	}

	/**
	 * Is called on an interval - figures out how many of the resources have loaded
	 * Called by: go
	 */
	this.initLoop = null;
	this.init = function() {

		// Find out how many image resources are finished
		var imagesLoaded = 0;
		for (var i in this.resources)
			if(this.resources[i].width > 0)
				imagesLoaded++;

		this.debug("I have loaded " + imagesLoaded + " of " + this.totalResources + " sprite" + (imagesLoaded != 1 ? 's' : '') + " and " + this.audioReady + " of " + this.totalAudio + " audio.");

		var p1 = this.getSprite("p1");
		this.updateCamera(p1.x - (this.gs.width / 2), p1.y - (this.gs.width / 2));

		// We're done loading sprites, move on to the logic loop
		if (imagesLoaded == this.totalResources && this.audioReady == this.totalAudio) {
			/**
			 * This is THE loop for ticking through the game steps! Don't mess it up :)
			 */
			clearInterval(this.initLoop);
			clearInterval(this.logicLoop);
			this.logicLoop = setInterval(function() {
				self.logic();
			}, this.gs.fps);
		}

		return;
	};//end function init

	/**
	 * Is called on an interval - performs the main game sequence.
	 * Called by: init
	 */
	this.logicCalls = 0;
	this.logicLoop = this.logicStart = this.logicTime = this.collideTime = null;
	this.collideChecks = 0;
	this.logic = function() {
		if(this.pause)
			return;
		this.logicCalls++;

		if (this.gs.showDebugging)
			this.logicStart = new Date();

		// Run the logic loop on all applicable sprites (non-null)
		var collide = [];
		for (var i in this.sprites) {
			for (j in this.sprites[i]) { 
				if (this.sprites[i][j] != null) {
					var s = this.sprites[i][j];
					s.renderOutline = false;

					// Only do logic for sprites which are on or around the viewable gamescreen
					var p = s.getRenderPoint(this.gs);
					s.renderX = p.x;
					s.renderY = p.y;

					// For X collisions, we care about the left and right edge
					if ((p.x + s.collideOffsetX) > (this.gs.width + 200))
						continue;
					if ((p.x + s.collideOffsetX + s.collideWidth) < -200)
						continue;
					
					// Collisions on the Y axis are interesting. Right now the game takes a vertical slice of the entire game screen,
					// top to bottom. This avoids having enemies fall through floors when dropping down to them.
//					if ((p.y + s.collideOffsetY > (this.gs.height) + 200)
//						continue;
//					if ((p.y + s.collideOffsetY + s.collideHeight) < -200)
//						continue;


					// This is definitely on the game screen - perform the logic step
					s.logic(this);

					// Place this in the pool for colliding
					if (i == "game") {
						if(this.gs.showDebugging)
							s.renderOutline = true;
						collide.push(s);
					}
				}
			}
		}

		// Here's a new approach for collision detection. So long as there is ANY collision, resolve the colliding sprites.
		if (this.gs.showDebugging) {
			this.collideChecks = 0;
			var d1 = new Date();
		}

		for(var i in collide) {
			// Find out if anything is colliding with us
			var s = collide[i];
			var collisions = s.checkCollision(collide);

			if(this.gs.showDebugging)
				this.collideChecks += collide.length;

			// This will handle the collision on a case-by-case basis
			for (var j in collisions)
				s.collide(this, collisions[j]);
		}

		if (this.gs.showDebugging) {
			var d2 = new Date();
			this.collideTime = (d2 - d1);
		}

		// Ramp up the game worldSpeed
		this.hammerTimeRecover();
		if(this.gs.keyA)
			this.hammerTime();

		if (this.gs.showDebugging)
			this.logicTime = new Date() - this.logicStart;

		if(this.logicCalls == 1)
			$(this.canvas).css({
				background: this.sceneSettings.background
			});

		// Finally, check and see if we're due for a checkpoint or restore
		if(this.flaggedForCheckpoint)
			this._checkpoint();
		else if(this.flaggedForRestore)
			this._restoreCheckpoint();

		// Fire off the render step
		this.render();
	};//end function logic



	/**
	 * Is indirectly called on an interval. Every time logic() is called it makes a call to render() as well
	 * Called by: logic
	 */
	this.renderTime = null;
	this.renderedSprites = 0;
	this.render = function() {

		// Help calculate how long it takes to render
		if (this.gs.showDebugging) {
			this.renderStart = new Date();
			this.renderedSprites = 0;
		}

		// Clear the screen
		this.context.clearRect(0, 0, this.gs.width, this.gs.height);

		if(this.gs.cameraShake > 0) {
			var x = this.gs.cameraX + ( ( Math.random() * this.gs.cameraShake ) + 1 - (this.gs.cameraShake / 2) );
			var y = this.gs.cameraY + ( ( Math.random() * this.gs.cameraShake ) + 1 - (this.gs.cameraShake / 2) );
			this.updateCamera(x, y);
		}
		this.cameraShakeRecover();

		// Render each sprite
		for (i in this.sprites) {
			for (j in this.sprites[i]) {
				if (this.sprites[i][j] != null) {
					/**
					 * This is taken care of in chrome, but I suspect that Firefox doesn't know to avoid rendering objects which are too far off screen
					 */
					var s = this.sprites[i][j];
					var renderPoint = s.getRenderPoint(this.gs);

					// Bail if the sprite is too far off screen. Start testing right, then left (most likely), then bottom and top
					if(renderPoint.x > this.gs.width)
						continue;
					if(renderPoint.x + s.renderWidth < 0)
						continue;
					if(renderPoint.y > this.gs.height)
						continue;
					if(renderPoint.y + s.renderHeight < 0)
						continue;

					// If we made it this far, render that sprite!
					if(this.gs.showDebugging)
						this.renderedSprites++;

					var suffix = (typeof this.sprites[i][j].currentAnimation == "string") ? "_" + this.sprites[i][j].currentAnimation : "";
					this.sprites[i][j].render(this.resources[this.sprites[i][j].name + suffix], this.context, this.gs, renderPoint);

					if (typeof s.renderOutline != "undefined" && s.renderOutline == true)
						s.outline(this.context, this.gs);
				}
			}
		}


		if (this.gs.showMeter){
			this.context.fillStyle = "rgba(255, 255, 255, 0.8)";
			this.context.fillRect(320, 0, 150, 15);
			this.context.fillStyle = "#000000";
			this.context.fillText("Fuel: " + this.gs.fuel, 375, 10)

		}
		

		// Render the FPS info
		if (this.gs.showDebugging) {
			var d = new Date();
			this.renderTime = d - this.renderStart;
			totalTime = d - this.logicStart;

			this.context.fillStyle = "rgba(255, 255, 255, 0.8)";
			this.context.fillRect(0, 0, this.gs.width, 15);

			this.context.fillStyle = "#000000";
			this.context.fillText("Total Sprites: " + this.countSprites(), 10, 10)
			this.context.fillText("Rendered: " + this.renderedSprites, 130, 10)
			this.context.fillText("Collide Checks: " + this.collideChecks, 230, 10)
			this.context.fillText("Gravity: " + this.gs.gravity, 350, 10)
			this.context.fillText("FPS: " + Math.round(1000 / totalTime), 430, 10);
			this.context.fillText("Logic Time: " + this.logicTime + "ms", 490, 10);
			this.context.fillText("Render Time: " + this.renderTime + "ms", 590, 10);
			this.context.fillText("Total Frames: " + this.logicCalls, 740, 10);
		}
	};//end function render

	// Attempts to update the camera and point it somewhere else
	this.updateCamera = function(x, y) {
		this.gs.cameraX = x;
		this.gs.cameraY = y;

		if(this.gs.cameraX < this.gs.cameraMinX)
			this.gs.cameraX = this.gs.cameraMinX;
		if(this.gs.cameraX > this.gs.cameraMaxX)
			this.gs.cameraX = this.gs.cameraMaxX;

		if(this.gs.cameraY < this.gs.cameraMinY)
			this.gs.cameraY = this.gs.cameraMinY;
		if(this.gs.cameraY > this.gs.cameraMaxY)
			this.gs.cameraY = this.gs.cameraMaxY;

	};//end function updateCamera

	/**
	 * Will attempt to play an audio resource as defined in declareScene. Will silently fail if the resource is non-existent.
	 */
	this.playAudio = function(audioResource) {
		var a = this.audio[audioResource];
		if(typeof a == "undefined")
			return;
		if(a.currentTime > 0){
			a.pause();
			a.currentTime = 0;
		}

		a.play();
	};//end function playAudio

	this.pauseAudio = function(audioResource) {
		if(typeof this.audio[audioResource] == "undefined")
			return;
		if(this.audio[audioResource].currentTime == 0)
			return;

		this.audio[audioResource].pause();
	};//end function stopAudio

	// Mapping for mouse/keypress functions
	$(function() {
		$(document).bind("keydown", function(event) {
			return self.gs.mapKeyEvent(true, event);
		}).bind("keyup", function(event) {
			return self.gs.mapKeyEvent(false, event);
		});
		var mouseTypes = ["mousedown", "mousemove", "mouseup"];
		for(var i in mouseTypes)
			$(self.canvas).unbind(mouseTypes[i]).bind(mouseTypes[i], function(event){
				return self.gs.mapMouseEvent(event);
			});
	});

	this.hammerTime = function() {
		// Only engage hammer time if the game is currently running at normal speeds
		if (this.gs.worldSpeed == this.gs.originalWorldSpeed) {
//			this.gs.worldSpeed = this.gs.worldSpeedMinimum;
//			this.gs.gravity *= this.gs.worldSpeed;
		}
	};//end function hammerTime

	this.cameraShakeRecover = function() {
		if(this.gs.cameraShake > 0) {
			this.gs.cameraShake /= this.gs.cameraShakeRamp;
			if(this.gs.cameraShake <= 1)
				this.gs.cameraShake = 0;
		}
	};

	this.hammerTimeRecover = function() {
		if(this.gs.worldSpeed < this.gs.originalWorldSpeed) {
			this.gs.worldSpeed *= this.gs.worldSpeedRamp;
			this.gs.gravity = this.gs.originalGravity * this.gs.worldSpeed;

			if (this.gs.worldSpeed >= this.gs.originalWorldSpeed) {
				this.gs.worldSpeed = this.gs.originalWorldSpeed;
//				this.gs.gravity = this.gs.originalGravity;
			}
		}
	};//end function hammerTimeRecover

	/**
	 * Performs a checkpoint operation and saves the state of all sprites
	 */
	this.flaggedForCheckpoint = false;
	this.spriteBackup = {};
	this.cameraBackup = {};
	this.intervalBackup = [];
	this.checkpoint = function() {
		this.flaggedForCheckpoint = true;
	}
	this._checkpoint = function(keepPaused) {
		this.pause = true;

		// Clean out any remaining checkpoints
		this.spriteBackup = {};
		this.cameraBackup = {
			x: this.gs.cameraX,
			y: this.gs.cameraY
		};

		// Back up the full sprite state.
		for(var layer in this.sprites) {
			this.spriteBackup[layer] = []
			for(var j in this.sprites[layer]) {
				// Magic. This does a deep copy of the sprite object and stores it away into the backup object
				var clone = $.extend(true, {}, this.sprites[layer][j]);
				if(clone.uniqueID == "p1")
					clone.speedX = 0;
				this.spriteBackup[layer].push(clone);
			}
		}

		// Back up the intevals
		for (var index in this.intervals) {
			this.intervalBackup.push({
				callback: this.intervals[index].callback,
				timeout: this.intervals[index].timeout
			});
		}

		if(keepPaused === true)
			return;

		this.pause = this.flaggedForCheckpoint = false;
	};
	this.restoreCheckpoint = function() {
		this.flaggedForRestore = true;
	};
	this._restoreCheckpoint = function() {
		this.pause = true;

		// Start from scratch on the engine sprites. Clean them out!
		this.sprites = {};
		this.gs.cameraX = this.cameraBackup.x;
		this.gs.cameraY = this.cameraBackup.y;

		// Restore the full backup of the sprites (then delete the backup)
		for(var layer in this.spriteBackup) {
			this.sprites[layer] = [];
			for(var j in this.spriteBackup[layer]) {
				var clone = this.spriteBackup[layer][j];
				this.sprites[layer].push(clone);
			}
		}

		// Restore the intervals in two steps. First, clear out all running intervals.
		if(this.intervals.length > 0)
			for (var index in this.intervals)
				clearInterval(this.intervals[index].interval);
		this.intervals = [];

		// Then re-initialize them from backup
		if (this.intervalBackup.length > 0)
			for (var index in this.intervalBackup)
				this.intervals.push({
					callback: this.intervalBackup[index].callback,
					timeout: this.intervalBackup[index].timeout,
					interval: setInterval(this.intervalBackup[index].callback, this.intervalBackup[index].timeout)
				});
		this.intervalBackup = [];

		this._checkpoint(true);

		this.pause = this.flaggedForRestore = false;
	};

	/**
	 * Writes a debugging string to any element with the attribute: id="debug"
	 */
	this.debug = function debug(message) {
		if(!this.debugging)
			return;

		if($("#debug").length == 0)
			$("body").append('<div id="debug"></div>');

		$("#debug").append("<div>" + this.dateFormat(new Date()) + " > " + message + "</div>");
		if($("#debug div").length > 10)
			$("#debug div:first").remove();
	};//end function debug

 	/**
	 * Returns an well-formatted date object in HH:ii:ss.mmmm format 
	 */
	this.dateFormat = function(date) {
		var H = date.getHours();
		if(H < 10)
			H = "0" + H;

		var i = date.getMinutes();
		if(i < 10)
			i = "0" + i;

		var s = date.getSeconds();
		if(s < 10)
			s = "0" + s;

		var m = date.getMilliseconds();
		if(m < 10)
			m = "000" + m;
		else if (m < 100)
			m = "00" + m;
		else if (m < 1000)
			m = "0" + m;

		return H + ":" + i + ":" + s + "." + m;
	}//end function dateFormat

	/**
	 * state: Returns the internal state of the engine as a string
	 */
	this.state = function() {
		var totalSprites = this.countSprites();
		var running = 'up ' + Math.round((new Date() - this.initTimestamp) / 1000 / 60) + ' minute' + (running != 1 ? 's' : '') 
		this.debug("Engine state requested. Engine was started at " + this.dateFormat(this.initTimestamp) + " (" + running + ")");
		this.debug("&nbsp;&nbsp;&nbsp;Width: " + this.gs.width + ", Height: " + this.gs.height + ", Logic Calls: " + this.logicCalls);
		this.debug("&nbsp;&nbsp;&nbsp;Non-Colliding Sprites: " + (totalSprites - this.sprites.game.length) + " Colliding Sprites: " + this.sprites.game.length);
	};//end function state

}//end object

window.e = new engine();

})(window);
