/**
 * Represents a sprite object - any 2d graphical element
 * @author Torgie
 */
function sprite(p) {
	// Avoid errors by creating an empty object if "p" is undefined
	if(typeof p == "undefined")
		p = {};

	this.type = p.type;
	this.name = p.name;
	this.src = p.src;

	this.x = p.x;
	this.y = p.y;
	this.width = p.width > 0 ? p.width : 50;
	this.height = p.height > 0 ? p.height : 50;

	this.renderWidth = (typeof p.renderWidth == "number") ? p.renderWidth : this.width;
	this.renderHeight = (typeof p.renderHeight == "number") ? p.renderHeight : this.height;

	// Set up the collide box, which in most cases will be {0, 0, renderWidth, renderHeight}
	this.collideOffsetX = (typeof p.collideOffsetX == "number") ? p.collideOffsetX : 0;
	this.collideOffsetY = (typeof p.collideOffsetY == "number") ? p.collideOffsetY : 0;
	this.collideWidth = (typeof p.collideWidth == "number") ? p.collideWidth : this.renderWidth;
	this.collideHeight = (typeof p.collideHeight == "number") ? p.collideHeight : this.renderHeight;

	this.currentAnimation = (typeof p.currentAnimation == "string") ? p.currentAnimation : void(0);

	// Optional logic declarations
	this.speedX = (typeof p.speedX == "number") ? p.speedX : 0;
	this.speedY = (typeof p.speedY == "number") ? p.speedY : 0;

	this.uniqueID = (typeof p.uniqueID == "string") ? p.uniqueID : void(0);
	this.specialID = (typeof p.specialID == "string") ? p.specialID : void(0);
	// Things in the background move more slowly
	this.speedChop = 1;

	// Tells the editor about any extra fields this sprite will need. Other hints can be pushed on as necessary
	this.editorHints = [{
		field: "uniqueID",
		type: "text",
		label: "Unique ID:"
	}];

	/**
	 * The logic function is always extended and always takes a gameState object as its only parameter
	 * @param {gameState} gs The engine's gameState object
	 */
	this.logic = function(engine) {
		
	};//end function logic

	this.collide = function(engine) {
		
	};//end function collide

	this.destroy = function(engine, i, j) {
		engine.sprites[i][j] = null;
	};

	// Resize the sprite. Most of the time this will remain unchanged, but sometimes resizing can be tricky!
	this.resize = function(width, height) {
		this.renderWidth = width;
		this.renderHeight = height;
	};

	/**
	 * The render function is called by the engine, and the customRender function must always be extended
	 */
	this.render = function(resource, context, gs, renderPoint) {
		if(!this.resourceReady(resource))
			return false;

		// Declare where the source x-position will be located. On non-animating sprites, this will always be 0
		var sourceX = 0;
		// For now, this will always be 0
		var sourceY = 0;

		// Offset the animation and deal with animation flipping if we animate
		if (typeof this.currentAnimation == "string") {
			var animation = this.animation[this.currentAnimation];

			// Roll over the animation every frame
			if (animation.flip > animation.flipEvery) {
				animation.currentFrame = (++animation.currentFrame % animation.frames);
				animation.flip = 0;
			}

			// Current frame, adjusted for left/right facing, or 0 if not applicable
			sourceX = (this.width * animation.currentFrame);
			if(typeof this.facingRight == "boolean")
				sourceX = sourceX + (!this.facingRight ? (this.width * animation.frames) : 0);
				
			// Current frame in the y directions, or 0 if not applicable
			if (this.specialID == "lunarLander")
				sourceY = (this.height * animation.currentFrameY);
			}
			
			
			if(this.width == 0 && this.height == 0)
				throw "Destination width/height was 0 for " + this.name;

		// Draw the sprite to the screen
		try {
			context.drawImage(
				resource,			// Resource object
				sourceX,			// Source X position
				sourceY,			// Source Y position
				this.width,			// Source width
				this.height,		// Source height
				renderPoint.x,		// Target X position
				renderPoint.y,		// Target Y position
				this.renderWidth,	// Target width
				this.renderHeight	// Target height
			);
		}
		catch(e) {
			console.debug(this.name + ".drawImage(" + resource + "," + sourceX + ",0," + this.width + "," + this.height + "," + renderPoint.x + "," + renderPoint.y + "," + this.width + "," + this.height + ")");
		}

		if(gs.showDebugging)
			this.debugRender(resource, context, gs, renderPoint);
	};//end function render

	this.debugRender = function(resource, context, gs, renderPoint) {

	};//end function debugRender

	/**
	 * A handy function for outlining sprites
	 */
	this.outlineWidth = 2;
	this.outline = function(context, gs) {
		// Get the x, y, width, height values
		var renderPoint = this.getRenderPoint(gs);
		var width  = this.renderWidth;
		var height = this.renderHeight;

		context.fillStyle = "#FF0000";
		var w = this.outlineWidth;
		// In order: top, right, bottom, left
		context.fillRect(renderPoint.x - w, renderPoint.y - w, width + (w * 2), w);
		// Right
		context.fillRect(renderPoint.x + width, renderPoint.y, w, height);
		// Bottom
		context.fillRect(renderPoint.x - w, renderPoint.y + height, width + (w * 2), w);
		// Left
		context.fillRect(renderPoint.x - w, renderPoint.y, w, height);
	};//end function outline

	/**
	 * Helper Functions - these are not always utilized by the sprite
	 */

	// Checks two sprites and tells if they are overlapping
	this.checkCollision = function(sprites) {
		var collisions = [];

		for(var i in sprites) {
			if(sprites[i] == null)
				continue;

			// Don't collide with ourselves!
			if(this.id == sprites[i].id)
				continue;

			var s = sprites[i];

			// Calculate the left, top, right and bottom of this sprite
			var l1 = this.x + this.collideOffsetX;
			var t1 = this.y + this.collideOffsetY;
			var r1 = l1     + this.collideWidth;
			var b1 = t1     + this.collideHeight;

			// Calculate the same for our target sprite
			var l2 = s.x + s.collideOffsetX;
			var t2 = s.y + s.collideOffsetY;
			var r2 = l2  + s.collideWidth;
			var b2 = t2  + s.collideHeight;

			// Case: Right edge does not overlap target left edge
		    if (r1 <= l2) continue;
			// Case: Left edge does not overlap target right edge
		    if (l1 >= r2) continue;
			// Case: Bottom edge does not overlap target top
		    if (b1 <= t2) continue;
			// Case: Top edge does not overlap target bottom
		    if (t1 >= b2) continue;

			// If *NONE* of those cases happened, it means that we're overlapping - push it on the stack
			collisions.push(s);
		}

		return collisions;
	};//end function checkCollision

	// ResourceReady checks to see if the resource has been loaded by the engine yet
	this.resourceReady = function(resource) {
		if (typeof resource == "object")
			if (typeof resource.width == "number")
				return (resource.width > 0);
	};//end function resourceReady

	// Swaps an animation for a new one
	this.setAnimation = function(newAnimation, offsets) {
		this.currentAnimation = newAnimation;
		this.currentFrame = 0;
		this.flip = 0;
	};//end function setAnimation

	// Calculates the proper X,Y coordinate based off the camera
	this.getRenderPoint = function(gs) {
		return {
			// Render the object further to the LEFT (negative X) relative to the camera
			x: Math.floor(this.x - (gs.cameraX * this.speedChop)),
			// Render the object further to the BOTTOM (positive Y) relative to the camera
			y: Math.floor(this.y - (gs.cameraY * this.speedChop))
		};
	};//end function getRenderPoint

	/**
	 * Resizes a sprite to a specific width and height
	 * @param {Object} width
	 * @param {Object} height
	 */
	this.resize = function(width, height) {
		this.renderWidth = width;
		this.renderHeight = height;
	};//end function resize

	// Called AUTOMATICALLY whenever this sprite is being printed out somewhere
	this.toString = function() {
		return this.id + ":" + this.name;
	};//end function toString

}//end object sprite

function sprite_invisible(p) {
	p.type = "invisible";
	p.src = "images/sprites/null.png";

	// Set the collide and render height as the passed (w,h), then overwrite it to (1,1)
	p.collideWidth = p.renderWidth = (typeof p.width == "number") ? p.width : 50;
	p.collideHeight = p.renderHeight = (typeof p.height == "number") ? p.height : 50;
	p.width = 1;
	p.height = 1;

	this.sprite = new sprite(p);

	// The color will be different depending on the type of sprite
	this.sprite.color = (typeof p.color != "undefined") ? p.color : "#FF0000"; 

	this.sprite.editorHints.push({
		field: "generateSize",
		type: "flag"
	});

	this.sprite.resize = function(width, height) {
		this.renderWidth = this.collideWidth = width;
		this.renderHeight = this.collideHeight = height;
	};

	this.sprite.debugRender = function(resource, context, gs, renderPoint) {
		context.fillStyle = this.color;
		context.fillRect(
			renderPoint.x,
			renderPoint.y,
			this.collideWidth,
			this.collideHeight
		);
		context.fillStyle = "#000000";
		context.fillText(this.name, renderPoint.x, renderPoint.y - 2);
	};

	return this.sprite;
}
